################################################################################
# psfmodel/gaussian.py
################################################################################

import numpy as np
from scipy.special import erf

from psfmodel import PSF


INV_SQRT_2 = 2**(-0.5)

class GaussianPSF(PSF):
    """A 2-D Gaussian symmetric PSF.

    The PSF can have different standard deviations in the X and Y directions.
    The standard deviations for X and Y can be locked up front when the
    GaussianPSF object is created or left to float so that future calls may
    specify them directly.

    Because these are so fast and easy to compute, we don't cache any results.
    """

    def __init__(self, sigma=None, mean=0., sigma_angle=0., movement=None,
                 movement_granularity=0.1,
                 sigma_x_range=(0.01,10.), sigma_y_range=(0.01,10.),
                 angle_subsample=7):
        """Create a GaussianPSF object describing a 2-D Gaussian PSF.

        Input:
            sigma     The standard deviation of the Gaussian. May be a scalar
                      in which case the value applies to both X and Y, or a
                      tuple (sigma_y,sigma_x) one of which may be None.
                      None for sigma or for sigma_x/y means that the sigma
                      will be supplied later.

            mean      The mean of the Gaussian. May be a scalar in which case
                      the value applies to both X and Y, or a tuple
                      (mean_y,mean_x).

            sigma_angle
                      The angle of the Gaussian. If None, angle is allowed to
                      float during optimization. angle ranges from 0 to pi, with
                      0 being "noon" (-Y) and pi/2 being "3" (+X) assuming that
                      (0,0) is in the top left corner.

            movement  The amount of motion blur in the (Y,X) direction. Must
                      be a tuple of scalars.

            movement_granularity
                      The number of pixels to step for each smear while doing
                      motion blur.

            sigma_x_range
                      The valid range for sigma_x if it is not specified
                      otherwise. This is used during PSF fitting to let
                      sigma_x float to its optimal value.

            sigma_y_range
                      The valid range for sigma_y if it is not specified
                      otherwise. This is used during PSF fitting to let
                      sigma_y float to its optimal value.

            angle_subsample
                      The amount of subsampling done during computation of the
                      2-D Gaussian integral when angle != 0.
        """

        PSF.__init__(self, movement, movement_granularity)

        if np.shape(sigma) == ():
            self.sigma_y = sigma
            self.sigma_x = sigma
        else:
            self.sigma_y, self.sigma_x = sigma
        if np.shape(mean) == ():
            self.mean_y = mean
            self.mean_x = mean
        else:
            self.mean_y, self.mean_x = mean
        self.sigma_angle = sigma_angle
        self.sigma_angle_subsample = angle_subsample

        if self.sigma_y is None:
            if sigma_y_range is not None:
                self.additional_params.append(tuple(sigma_y_range)+('sigma_y',))
        if self.sigma_x is None:
            if sigma_x_range is not None:
                self.additional_params.append(tuple(sigma_x_range)+('sigma_x',))
        if self.sigma_angle is None:
                self.additional_params.append((0, np.pi, 'sigma_angle'))

    @staticmethod
    def gaussian_1d(x, sigma=1., mean=0., scale=1.0, base=0.):
        """Return a 1-D Gaussian.

        This simply returns the value of the Gaussian at a series of points.
        The Gaussian is normalized so the area under the curve is "scale".

        Input:
            x          A scalar or array
            sigma      The standard deviation of the Gaussian
            mean       The mean of the Gaussian
            scale      The scale of the Gaussian; the area under the complete
                       curve (excluding the base)
            base       The base of the Gaussian; a scalar added to the curve
        """

        return (scale/np.sqrt(2*np.pi*sigma**2) *
                np.exp(-(x-mean)**2/(2*sigma**2))) + base

    @staticmethod
    def gaussian_2d_rho(y, x, sigma_y=1., sigma_x=1., mean_y=0., mean_x=0.,
                        scale=1., base=0., rho=0.):
        """Return a 2-D Gaussian using rho (X/Y correlation).

        This simply returns the value of the 2-D Gaussian at a series of points.
        The Gaussian is normalized so the area under the curve is scale.

        Input:
            y          A scalar or array
            x          A scalar or array

            sigma_y    The standard deviation of the Gaussian in the Y
                       direction
            sigma_x    The standard deviation of the Gaussian in the X
                       direction

            mean_y     The mean of the Gaussian in the Y dimension
            mean_x     The mean of the Gaussian in the X dimension

            scale      The scale of the Gaussian; the area under the complete
                       curve (excluding the base)
            base       The base of the Gaussian; a scalar added to the curve

            rho        The correlation between the X and Y dimensions
                       From gaussians-cern.pdf:
                           sigma_X^2 = cos^2 sigma_x^2 + sin^2 sigma_y^2
                           sigma_Y^2 = cos^2 sigma_y^2 + sin^2 sigma_x^2
                        where sigma_X/Y are the size of the ellipse projected
                        onto the normal Cartesian axes. Also:
                           (1-rho^2) sigma_X^2 sigma_Y^2 = sigma_x^2 sigma_y^2
        """

        # See http://cs229.stanford.edu/section/gaussians.pdf
        # See https://indico.cern.ch/category/6015/attachments/192/632/Statistics_Gaussian_I.pdf
        x = x-mean_x
        y = y-mean_y

        norm_fact = 1./(2*np.pi*np.sqrt((sigma_x**2 * sigma_y**2 - rho**2)))
        expon = (((x*sigma_y)**2 + (y*sigma_x)**2 - 2*x*y*xcorr) /
                 ((sigma_x*sigma_y)**2-2*xcorr))
        return scale * norm_fact * np.exp(-0.5 * expon) + base

    @staticmethod
    def gaussian_2d(y, x, sigma_y=1., sigma_x=1., mean_y=0., mean_x=0.,
                    scale=1., base=0., sigma_angle=0.):
        """Return a 2-D Gaussian using angle (angle 0-pi, 0 at noon, CW).

        This simply returns the value of the 2-D Gaussian at a series of points.
        The Gaussian is normalized so the area under the curve is scale.

        Input:
            y          A scalar or array
            x          A scalar or array

            sigma_y    The standard deviation of the Gaussian in the Y
                       direction
            sigma_x    The standard deviation of the Gaussian in the X
                       direction

            mean_y     The mean of the Gaussian in the Y dimension
            mean_x     The mean of the Gaussian in the X dimension

            scale      The scale of the Gaussian; the area under the complete
                       curve (excluding the base)
            base       The base of the Gaussian; a scalar added to the curve

            sigma_angle
                       The angle of the ellipse in the range (0,pi). angle
                       ranges from 0 to pi, with 0 being "noon" (-Y) and pi/2
                       being "3" (+X) assuming that (0,0) is in the top left corner.
        """

        # See http://cs229.stanford.edu/section/gaussians.pdf
        # See https://indico.cern.ch/category/6015/attachments/192/632/Statistics_Gaussian_I.pdf

        # x and y, and sigma_x and sigma_y are in the rotated coordinate system
        x = x-mean_x
        y = y-mean_y

        # Convert angle (normally 0 is to the right, CW) to clock angle
        # (0 is noon, pi/2 is 3, pi is 6)
        sigma_angle = sigma_angle - np.pi/2

        # Convert x and y (ellipse coordinates) to X and Y (Cartesian
        # coordinates)
        c = np.cos(sigma_angle)
        s = np.sin(sigma_angle)
        X = c*x + s*y
        Y = -s*x + c*y

        norm_fact = 1./(2*np.pi*np.sqrt((sigma_x**2 * sigma_y**2)))
        expon = (((X*sigma_y)**2 + (Y*sigma_x)**2) /
                 ((sigma_x*sigma_y)**2))
        return scale * norm_fact * np.exp(-0.5 * expon) + base

    @staticmethod
    def gaussian_integral_1d(xmin, xmax, sigma=1., mean=0.,
                             scale=1., base=0.):
        """Return the integral of a Gaussian.

        The integral is over the limits [xmin,xmax].

        Values are generated via the error function, where the integral from
        -inf to x is equal to
                    (1 + erf((x - mean_x) / (sqrt(2)*sigma_x)) / 2

        This function works for both scalar and array values of xmin and xmax.

        Input:
            xmin       The lower bound of the integral; may be a scalar or 1-D
                       array
            xmax       The upper bound of the integral; may be a scalar or 1-D
                       array
            sigma_x    The standard deviation of the Gaussian
            mean_x     The mean of the Gaussian
            scale      The scale of the Gaussian; the area under the complete
                       curve (excluding the base)
            base       The base of the Gaussian; a scalar added to the curve
        """

        # Normalize xmin and xmax
        assert sigma > 0.
        xmin_div_sqrt_2 = (xmin - mean) * (INV_SQRT_2/sigma)
        xmax_div_sqrt_2 = (xmax - mean) * (INV_SQRT_2/sigma)

        # Handle the scalar case
        if np.shape(xmin) == () and np.shape(xmax) == ():
            return (0.5 * (erf(xmax_div_sqrt_2) -
                           erf(xmin_div_sqrt_2)) * scale) + base

        # If either value is an array, broadcast to a common shape
        (xmin_div_sqrt_2,
         xmax_div_sqrt_2) = np.broadcast_arrays(xmin_div_sqrt_2,
                                                xmax_div_sqrt_2)

        result = erf(xmax_div_sqrt_2) - erf(xmin_div_sqrt_2)

        return result * 0.5 * scale + base

    @staticmethod
    def gaussian_integral_2d(ymin, ymax, xmin, xmax, sigma_y=1., sigma_x=1.,
                             mean_y=0., mean_x=0., scale=1., base=0.,
                             sigma_angle=0., sigma_angle_subsample=7):
        """Return the double integral of a 2-D Gaussian.

        The integral is over the limits [ymin,ymax] and [xmin,xmax].

        This function works for both scalar and array values of
        xmin/xmax/ymin/ymax.

        Inputs:
            ymin       The lower bound of the integral in the Y dimension;
                       may be a scalar or 1-D array
            ymax       The upper bound of the integral in the Y dimension;
                       may be a scalar or 1-D array
            xmin       The lower bound of the integral in the X dimension;
                       may be a scalar or 1-D array
            xmax       The upper bound of the integral in the X dimension;
                       may be a scalar or 1-D array

            sigma_y    The standard deviation of the Gaussian in the Y
                       dimension
            sigma_x    The standard deviation of the Gaussian in the X
                       dimension

            mean_y     The mean of the Gaussian in the Y dimension
            mean_x     The mean of the Gaussian in the X dimension

            scale      The scale of the Gaussian; the area under the complete
                       curve (excluding the base)
            base       The base of the Gaussian; a scalar added to the curve

            sigma_angle
                       The angle of the ellipse in the range (0,pi)

            sigma_angle_subsample
                      The amount of subsampling done during computation of the
                      2-D Gaussian integral when angle != 0.
        """

        if sigma_angle == 0.:
            return (GaussianPSF.gaussian_integral_1d(ymin, ymax,
                                                     sigma_y, mean_y) *
                    GaussianPSF.gaussian_integral_1d(xmin, xmax,
                                                     sigma_x, mean_x) *
                    scale + base)

        # Handle the scalar case
        if (np.shape(xmin) == () and np.shape(xmax) == () and
            np.shape(ymin) == () and np.shape(ymax) == ()):
            ys = np.linspace(ymin, ymax, sigma_angle_subsample)
            xs = np.linspace(xmin, xmax, sigma_angle_subsample)
            xindex, yindex = np.meshgrid(xs, ys)

            ret = GaussianPSF.gaussian_2d(ys, xs, sigma_y, sigma_x,
                                          mean_y, mean_x, scale, base,
                                          sigma_angle)
            return np.mean(ret)

        res = np.empty(xmin.shape)
        for x in range(xmin.shape[0]):
            ys = np.linspace(ymin[x], ymax[x], sigma_angle_subsample)
            xs = np.linspace(xmin[x], xmax[x], sigma_angle_subsample)
            xindex, yindex = np.meshgrid(xs, ys)

            ret = GaussianPSF.gaussian_2d(ys, xs, sigma_y, sigma_x,
                                          mean_y, mean_x, scale, base,
                                          sigma_angle)
            res[x] = np.mean(ret)

        return res
        # If any value is an array, broadcast to a common shape
#         (xmin, xmax, ymin, ymax) = np.broadcast_arrays(xmin, xmax,
#                                                        ymin, ymax)


    def eval_point(self, coord, scale=1., base=0., sigma=None,
                   sigma_y=None, sigma_x=None, sigma_angle=None):
        """Evaluate the 2-D Gaussian PSF at a single, fractional, point.

        (0,0) is the center of the PSF and x and y may be negative.

        Input:
            coord       The coordinate (y,x) at which to evaluate the PSF.

            scale       A scale factor to apply to the resulting PSF.
            base        A scalar added to the resulting PSF.

            sigma       The standard deviation of the Gaussian. It may be
                        specified here or during the creation of the
                        GaussianPSF object but not both. May be a scalar
                        or a tuple (sigma_y,sigma_x), or None if specified
                        at creation time.

            sigma_y     An alternative way to specify sigma_y. Used primarily
                        for letting sigma_y float during PSF fitting.
            sigma_x     An alternative way to specify sigma_x. Used primarily
                        for letting sigma_x float during PSF fitting.
            sigma_angle An alternative way to specify angle. Used primarily
                        for letting angle float during PSF fitting.
        """

        sy = self.sigma_y
        sx = self.sigma_x

        if sigma is not None:
            if np.shape(sigma) == ():
                sy = sigma
                sx = sigma
            else:
                sy, sx = sigma

        if sigma_y is not None:
            sy = sigma_y
        if sigma_x is not None:
            sx = sigma_x

        r = self.sigma_angle
        if sigma_angle is not None:
            r = sigma_angle

        return PSFGaussian.gaussian_2d(coord[0], coord[1], sy, sx,
                                       self.mean_y, self.mean_x,
                                       scale, base, r)

    def eval_pixel(self, coord, offset=(0.,0.), scale=1., base=0., sigma=None,
                   sigma_y=None, sigma_x=None, sigma_angle=None):
        """Evaluate the Gaussian PSF integrated over an entire integer pixel.

        The returned array has the PSF offset from the center by
        (offset_y,offset_x). An offset of (0,0) places the PSF in the upper
        left corner of the center pixel while an offset of (0.5,0.5)
        places the PSF in the center of the center pixel. The angle is applied
        relative to this new origin, so as angle changes the center of the
        ellipse does not move.

        This essentially performs a 2-D integration of the PSF over the intervals
            [y-offset_y-0.5,y-offset_y+0.5] and
            [x-offset_x-0.5,x-offset_x+0.5].

        Input:
            coord       The integer coordinate (y,x) at which to evaluate the
                        PSF.
            offset      The amount (offset_y,offset_x) to offset the center
                        of the PSF.

            scale       A scale factor to apply to the resulting PSF.
            base        A scalar added to the resulting PSF.

            sigma       The standard deviation of the Gaussian. It may be
                        specified here or during the creation of the
                        GaussianPSF object but not both. May be a scalar
                        or a tuple (sigma_y,sigma_x), or None if specified
                        at creation time.

            sigma_y     An alternative way to specify sigma_y. Used primary
                        for letting sigma_y float during astrometry.
            sigma_x     an alternative way to specify sigma_x. Used primary
                        for letting sigma_x float during astrometry.
            sigma_angle An alternative way to specify angle. Used primarily
                        for letting angle float during PSF fitting.
        """

        sy = self.sigma_y
        sx = self.sigma_x

        if sigma is not None:
            if np.shape(sigma) == ():
                sy = sigma
                sx = sigma
            else:
                sy, sx = sigma

        if sigma_y is not None:
            sy = sigma_y
        if sigma_x is not None:
            sx = sigma_x

        r = self.sigma_angle
        if sigma_angle is not None:
            r = sigma_angle

        return (GaussianPSF.gaussian_integral_2d(coord[0]-offset[0],
                                                 coord[0]-offset[0]+1.,
                                                 coord[1]-offset[1],
                                                 coord[1]-offset[1]+1,
                                                 sy, sx,
                                                 self.mean_y, self.mean_x,
                                                 scale, base, r,
                                                 self.sigma_angle_subsample))

    def _eval_rect(self, rect_size, offset=(0.,0.), scale=1., base=0.,
                   sigma=None, sigma_y=None, sigma_x=None, sigma_angle=None):
        rect_size_y, rect_size_x = rect_size
        y_coords = np.repeat(np.arange(-(rect_size_y//2), rect_size_y//2+1),
                             rect_size_x).astype('float')
        x_coords = np.tile(np.arange(-(rect_size_x//2), rect_size_x//2+1),
                           rect_size_y).astype('float')
        coords = np.empty((2,rect_size_y*rect_size_x))
        coords[0] = y_coords
        coords[1] = x_coords
        rect = self.eval_pixel(coords, offset, scale, base, sigma,
                               sigma_y, sigma_x, sigma_angle)
        rect = rect.reshape(rect_size)

        return rect

    def eval_rect(self, rect_size, offset=(0.,0.), scale=1., base=0.,
                  sigma=None, sigma_y=None, sigma_x=None, sigma_angle=None):
        """Create a rectangular pixelated Gaussian PSF.

        This is done by evaluating the PSF function from:
            [-rect_size_y//2:rect_size_y//2] and
            [-rect_size_x//2:rect_size_x//2]

        The returned array has the PSF offset from the center by
        (offset_y,offset_x). An offset of (0,0) places the PSF in the upper
        left corner of the center pixel while an offset of (0.5,0.5)
        places the PSF in the center of the center pixel. The angle is applied
        relative to this new origin, so as angle changes the center of the
        ellipse does not move.

        Input:
            rect_size   The size of the rectangle (rect_size_y,rect_size_x)
                        of the returned PSF. Both dimensions must be odd.
            offset      The amount (offset_y,offset_x) to offset the center
                        of the PSF from the upper left corner of the center
                        pixel of the rectangle.

            scale       A scale factor to apply to the resulting PSF.
            base        A scalar added to the resulting PSF.

            sigma       The standard deviation of the Gaussian. It may be
                        specified here or during the creation of the
                        GaussianPSF object but not both. May be a scalar
                        or a tuple (sigma_y,sigma_x), or None if specified
                        at creation time.

            sigma_y     An alternative way to specify sigma_y. Used primary
                        for letting sigma_y float during astrometry.
            sigma_x     An alternative way to specify sigma_x. Used primary
                        for letting sigma_x float during astrometry.
            sigma_angle An alternative way to specify angle. Used primarily
                        for letting angle float during PSF fitting.
        """

        rect_size_y, rect_size_x = rect_size
        offset_y, offset_x = offset

        assert rect_size_y % 2 == 1 # Odd
        assert rect_size_x % 2 == 1 # Odd

        if self.movement[0] == 0. and self.movement[1] == 0.:
            return self._eval_rect(rect_size, offset, scale, base, sigma,
                                   sigma_y, sigma_x, sigma_angle)

        return self._eval_rect_smeared(self.movement, rect_size, offset,
                                       scale, base, sigma, sigma_y, sigma_x,
                                       sigma_angle)



################################################################################
# UNIT TESTS
################################################################################

import unittest

class Test_GaussianPSF(unittest.TestCase):

    def runTest(self):

        # gaussian_1d
        x_coords = np.arange(-20., 21.)/4
        gauss1d = GaussianPSF.gaussian_1d(x_coords)/4
        self.assertEqual(np.argmax(gauss1d), 20)
        self.assertAlmostEqual(np.sum(gauss1d), 1., places=5)

        gauss1d_scalar = GaussianPSF.gaussian_1d(0.)/4
        self.assertEqual(gauss1d_scalar, gauss1d[20])

        gauss1d2 = GaussianPSF.gaussian_1d(x_coords, sigma=0.2)/4
        self.assertAlmostEqual(np.sum(gauss1d2), 1., places=4)
        self.assertTrue(np.sum(np.abs(gauss1d-gauss1d2)) > 1e-5)

        gauss1d3 = GaussianPSF.gaussian_1d(x_coords, mean=0.25)/4
        self.assertEqual(np.argmax(gauss1d3), 21)
        self.assertAlmostEqual(np.sum(gauss1d3), 1., places=5)
        self.assertTrue(np.sum(np.abs(gauss1d-gauss1d3)) > 1e-5)

        gauss1d4 = GaussianPSF.gaussian_1d(x_coords, scale=2.)/4
        self.assertAlmostEqual(np.sum(gauss1d4), 2., places=5)

        gauss1d5 = GaussianPSF.gaussian_1d(x_coords, base=1.)/4
        self.assertAlmostEqual(np.sum(gauss1d5-.25), 1., places=5)

        # gaussian_2d
        y_coords = np.tile(np.arange(-10., 11.)/2, 21)
        x_coords = np.repeat(np.arange(-10., 11.)/2, 21)
        gauss2d = GaussianPSF.gaussian_2d(y_coords, x_coords)/4
        self.assertAlmostEqual(np.sum(gauss2d), 1., places=5)

        gauss2d_scalar = GaussianPSF.gaussian_2d(0., 0.)/4
        self.assertEqual(gauss2d_scalar, gauss2d[10*21+10])

        gauss2d = GaussianPSF.gaussian_2d(y_coords, x_coords, scale=2.)/4
        self.assertAlmostEqual(np.sum(gauss2d), 2., places=5)

        gauss2d5 = GaussianPSF.gaussian_2d(y_coords, x_coords, scale=2.,
                                           sigma_x=.5, sigma_y=.5)/4
        self.assertAlmostEqual(np.sum(gauss2d5), 2., places=5)
        self.assertTrue(np.sum(np.abs(gauss2d-gauss2d5)) > 1e-5)

        gauss2d51 = GaussianPSF.gaussian_2d(y_coords, x_coords, scale=2.,
                                           sigma_x=.5, sigma_y=.5,
                                           base=1.)/4
        self.assertAlmostEqual(np.sum(gauss2d51-0.25), 2., places=5)
        self.assertTrue(np.sum(np.abs(gauss2d51-gauss2d5)) > 1e-5)

        gauss2d2 = GaussianPSF.gaussian_2d(y_coords, x_coords, scale=2.,
                                           sigma_x=.25, sigma_y=.5,
                                           base=1.)/4
        gauss2d3 = GaussianPSF.gaussian_2d(y_coords, x_coords, scale=2.,
                                           sigma_x=.5, sigma_y=.25,
                                           base=1.)/4
        gauss2d2 = gauss2d2.reshape(21,21)
        gauss2d3 = gauss2d3.reshape(21,21)
        self.assertTrue(np.sum(np.abs(gauss2d2-gauss2d3)) > 1e-5)
        self.assertTrue(np.sum(np.abs(np.transpose(gauss2d2)-gauss2d3)) < 1e-5)

        # gaussian_integral_1d
        i1 = GaussianPSF.gaussian_integral_1d(-10., 10.)
        self.assertAlmostEqual(i1, 1.)
        i2 = GaussianPSF.gaussian_integral_1d(-10., 0.)
        i3 = GaussianPSF.gaussian_integral_1d(0., 10.)
        self.assertAlmostEqual(i2, i3)
        i4 = GaussianPSF.gaussian_integral_1d(0., 10., sigma=5.)
        self.assertNotAlmostEqual(i2, i4)
        i5 = GaussianPSF.gaussian_integral_1d(0., 10., mean=1.)
        self.assertNotAlmostEqual(i2, i5)
        i6 = GaussianPSF.gaussian_integral_1d(-10., 1., mean=1., sigma=0.25)
        i7 = GaussianPSF.gaussian_integral_1d(1., 10., mean=1., sigma=0.25)
        self.assertAlmostEqual(i6, i7)
        i8 = GaussianPSF.gaussian_integral_1d(-10., 10., scale=2., base=1.)
        self.assertAlmostEqual(i8, 3.)

        # NEED TO TEST ARRAY VERSION

        # gaussian_integral_2d
        i1 = GaussianPSF.gaussian_integral_2d(-10., 10., -10., 10.)
        self.assertAlmostEqual(i1, 1.)
        i2 = GaussianPSF.gaussian_integral_2d(-10., 0., -10., 10.)
        i3 = GaussianPSF.gaussian_integral_2d(0., 10., -10., 10.)
        self.assertAlmostEqual(i2, i3)
        i2 = GaussianPSF.gaussian_integral_2d(-10., 10., -10., 0.)
        i3 = GaussianPSF.gaussian_integral_2d(-10., 10., 0., 10.)
        self.assertAlmostEqual(i2, i3)
        i4 = GaussianPSF.gaussian_integral_2d(-10., 10., -10., 0., sigma_x=5.)
        self.assertNotAlmostEqual(i2, i4)
        i4 = GaussianPSF.gaussian_integral_2d(-10., 10., -10., 0., sigma_y=5.)
        self.assertNotAlmostEqual(i2, i4)
        i5 = GaussianPSF.gaussian_integral_2d(-10., 10., -10., 0., mean_x=1.)
        self.assertNotAlmostEqual(i2, i5)
        i5 = GaussianPSF.gaussian_integral_2d(-10., 10., -10., 0., mean_y=10.)
        self.assertNotAlmostEqual(i2, i5)
        i6 = GaussianPSF.gaussian_integral_2d(-10., 0., -10., 10.,
                                              sigma_x=0.5, sigma_y=2.0)
        i7 = GaussianPSF.gaussian_integral_2d(0., 10., -10., 10.,
                                              sigma_x=0.5, sigma_y=2.0)
        self.assertAlmostEqual(i6, i7)
        i8 = GaussianPSF.gaussian_integral_2d(-10., 10., -10., 10.,
                                              scale=2., base=1.)
        self.assertAlmostEqual(i8, 3.)

        # eval_point

        # eval_rect

        # eval_pixel and optimize_from_data
        psf = GaussianPSF()
        y_coords = np.repeat(np.arange(-10., 11.), 21)
        x_coords = np.tile(np.arange(-10., 11.), 21)
        coords = np.empty((2,21*21))
        coords[0] = y_coords
        coords[1] = x_coords
        gauss2d = psf.eval_pixel(coords, scale=2., sigma=1.)
        gauss2d = gauss2d.reshape(21,21)
        ret = psf.find_position(gauss2d, gauss2d.shape,
                                starting_point=((gauss2d.shape[0]//2,
                                                 gauss2d.shape[1]//2)),
                                bkgnd_degree=0, num_sigma=0)
        self.assertAlmostEqual(ret[0], gauss2d.shape[0]//2)
        self.assertAlmostEqual(ret[1], gauss2d.shape[1]//2)
        self.assertAlmostEqual(ret[2]['sigma_y'], 1.)
        self.assertAlmostEqual(ret[2]['sigma_x'], 1.)
        self.assertAlmostEqual(ret[2]['scale'], 2., places=6)

        y_coords = np.repeat(np.arange(-10., 11.), 21)
        x_coords = np.tile(np.arange(-10., 11.), 21)
        coords = np.empty((2,21*21))
        coords[0] = y_coords
        coords[1] = x_coords
        gauss2d = psf.eval_pixel(coords, scale=2., sigma=(2., 0.5))
        gauss2d = gauss2d.reshape(21,21)
        ret = psf.find_position(gauss2d, gauss2d.shape,
                                starting_point=((gauss2d.shape[0]//2,
                                                 gauss2d.shape[1]//2)),
                                bkgnd_degree=0, num_sigma=0)
        self.assertAlmostEqual(ret[0], gauss2d.shape[0]//2)
        self.assertAlmostEqual(ret[1], gauss2d.shape[1]//2)
        self.assertAlmostEqual(ret[2]['sigma_y'], 2.)
        self.assertAlmostEqual(ret[2]['sigma_x'], 0.5)
        self.assertAlmostEqual(ret[2]['scale'], 2., places=6)

        psf2 = GaussianPSF(mean=(0.5,0.75))
        y_coords = np.repeat(np.arange(-10., 11.), 21)
        x_coords = np.tile(np.arange(-10., 11.), 21)
        coords = np.empty((2,21*21))
        coords[0] = y_coords
        coords[1] = x_coords
        gauss2d = psf2.eval_pixel(coords, scale=0.5, sigma=(0.3, 1.3))
        gauss2d = gauss2d.reshape(21,21)

        ret = psf.find_position(gauss2d, gauss2d.shape,
                                starting_point=((gauss2d.shape[0]//2,
                                                 gauss2d.shape[1]//2)),
                                bkgnd_degree=0, num_sigma=0)
        self.assertAlmostEqual(ret[0], gauss2d.shape[0]//2+0.5)
        self.assertAlmostEqual(ret[1], gauss2d.shape[1]//2+0.75)
        self.assertAlmostEqual(ret[2]['sigma_y'], 0.3)
        self.assertAlmostEqual(ret[2]['sigma_x'], 1.3)
        self.assertAlmostEqual(ret[2]['scale'], 0.5, places=6)

        ret = psf2.find_position(gauss2d, gauss2d.shape,
                                 starting_point=((gauss2d.shape[0]//2,
                                                  gauss2d.shape[1]//2)),
                                 bkgnd_degree=0, num_sigma=0)
        self.assertAlmostEqual(ret[0], gauss2d.shape[0]//2)
        self.assertAlmostEqual(ret[1], gauss2d.shape[1]//2)
        self.assertAlmostEqual(ret[2]['sigma_y'], 0.3)
        self.assertAlmostEqual(ret[2]['sigma_x'], 1.3)
        self.assertAlmostEqual(ret[2]['scale'], 0.5, places=6)

        psf2 = GaussianPSF()
        y_coords = np.repeat(np.arange(-10., 11.), 21)
        x_coords = np.tile(np.arange(-10., 11.), 21)
        coords = np.empty((2,21*21))
        coords[0] = y_coords
        coords[1] = x_coords
        gauss2d = psf2.eval_pixel(coords, offset=(0.21, -0.35),
                                  scale=0.5, sigma=(0.3, 1.3))
        gauss2d = gauss2d.reshape(21,21)
        ret = psf.find_position(gauss2d, gauss2d.shape,
                                starting_point=((gauss2d.shape[0]//2,
                                                 gauss2d.shape[1]//2)),
                                bkgnd_degree=0, num_sigma=0)
        self.assertAlmostEqual(ret[0], gauss2d.shape[0]//2+0.21)
        self.assertAlmostEqual(ret[1], gauss2d.shape[1]//2-0.35)
        self.assertAlmostEqual(ret[2]['sigma_y'], 0.3)
        self.assertAlmostEqual(ret[2]['sigma_x'], 1.3)
        self.assertAlmostEqual(ret[2]['scale'], 0.5, places=6)


########################################
if __name__ == '__main__':
    unittest.main(verbosity=2)
################################################################################



################################################################################
# psfmodel/hst.py
################################################################################

import os

import astropy.io.fits as pyfits
import numpy as np
from scipy.interpolate import RectBivariateSpline
import scipy.signal as scisig

from psfmodel import PSF


#===============================================================================
#
# HST CAMERA DETAILS
#
#===============================================================================

#=====================
# ACS HRC is 1024x1024
#=====================

# TinyTim returns the following size PSFs given the FOV in arcsec.
# Subsample > 1 may make the returned PSF slightly larger than predicted
ACS_HRC_RETURNED_SIZES = [
    (0.5,  27),
    (0.75, 39),
    (1.0,  50),
    (1.25, 62),
    (1.5,  75),
    (1.75, 87),
    (2.0,  98),
    (2.25, 111),
    (2.5,  123),
    (2.75, 136),
    (3.0,  146),
    (4.0,  194),
    (5.0,  243),
    (6.0,  291)
]

ACS_HRC_PIXEL_SIZE = 0.025 # arc sec
ACS_HRC_DEFAULT_RECTSIZEGUESS = 11 # Safely inside 0.75 arcsec
ACS_HRC_DEFAULT_FOV = 0.75 # arc sec
ACS_HRC_DEFAULT_JITTER = 10
ACS_HRC_DEFAULT_SUBSAMPLE = 5
ACS_HRC_DEFAULT_LINE = 512
ACS_HRC_DEFAULT_SAMPLE = 512

#=====================
# WFPC2 PC1 is 800x800
#=====================

# TinyTim returns the following size PSFs given the FOV in arcsec.
# Subsample > 1 may make the returned PSF slightly larger than predicted
WFPC2_PC1_RETURNED_SIZES = [
    (0.5,  10),
    (0.75, 16),
    (1.0,  20),
    (1.25, 26),
    (1.5,  32),
    (1.75, 38),
    (2.0,  42),
    (2.25, 48),
    (2.5,  54),
    (2.75, 60),
    (3.0,  64),
    (4.0,  86),
    (5.0,  108),
    (6.0,  130)
]

WFPC2_PC1_PIXEL_SIZE = 0.046 # arc sec
WFPC2_PC1_DEFAULT_RECTSIZEGUESS = 11 # Safely inside 0.75 arcsec
WFPC2_PC1_DEFAULT_FOV = 0.75 # arc sec
WFPC2_PC1_DEFAULT_JITTER = 20
WFPC2_PC1_DEFAULT_SUBSAMPLE = 5
WFPC2_PC1_DEFAULT_LINE = 400
WFPC2_PC1_DEFAULT_SAMPLE = 400

#=======================
# WFC3 IR is 1014x1014
# WFC3 UVIS is 4096x2051
#=======================

# TinyTim returns the following size PSFs given the FOV in arcsec.
# Subsample > 1 may make the returned PSF slightly larger than predicted
WFC3_UVIS_RETURNED_SIZES = [
    (0.5,  22),
    (0.75, 32),
    (1.0,  28),
    (1.25, 36),
    (1.5,  44),
    (1.75, 51),
    (2.0,  59),
    (2.25, 66),
    (2.5,  74),
    (2.75, 81),
    (3.0,  89),
    (4.0,  119),
    (5.0,  149),
    (6.0,  179)
]

WFC3_UVIS_PIXEL_SIZE = 0.0396 # arc sec
WFC3_UVIS_DEFAULT_RECTSIZEGUESS = 11 # Safely inside 0.75 arcsec
WFC3_UVIS_DEFAULT_FOV = 0.75 # arc sec
WFC3_UVIS_DEFAULT_JITTER = 10
WFC3_UVIS_DEFAULT_SUBSAMPLE = 5
WFC3_UVIS_DEFAULT_LINE = None
WFC3_UVIS_DEFAULT_SAMPLE = None

# TinyTim returns the following size PSFs given the FOV in arcsec.
# Subsample > 1 may make the returned PSF slightly larger than predicted
WFC3_IR_RETURNED_SIZES = [ # XXX
    (0.5,  4),
    (0.75, 6),
    (1.0,  9),
    (1.25, 11),
    (1.5,  13),
    (1.75, 15),
    (2.0,  18),
    (2.25, 20),
    (2.5,  22),
    (2.75, 24),
    (3.0,  27),
    (4.0,  36),
    (5.0,  45),
    (6.0,  54)
]

WFC3_IR_PIXEL_SIZE = 0.128 # arc sec
WFC3_IR_DEFAULT_RECTSIZEGUESS = 11 # Safely inside 2.0 arcsec
WFC3_IR_DEFAULT_FOV = 2.0 # arc sec
WFC3_IR_DEFAULT_JITTER = 40
WFC3_IR_DEFAULT_SUBSAMPLE = 5
WFC3_IR_DEFAULT_LINE = None
WFC3_IR_DEFAULT_SAMPLE = None


RETURNED_SIZES = {
    ('ACS',   'HRC'):  ACS_HRC_RETURNED_SIZES,
    ('WFPC2', 'PC1'):  WFPC2_PC1_RETURNED_SIZES,
    ('WFC3',  'UVIS'): WFC3_UVIS_RETURNED_SIZES,
    ('WFC3',  'IR'):   WFC3_IR_RETURNED_SIZES,
}

DEFAULT_RECTSIZEGUESS = {
    ('ACS',   'HRC'):  ACS_HRC_DEFAULT_RECTSIZEGUESS,
    ('WFPC2', 'PC1'):  WFPC2_PC1_DEFAULT_RECTSIZEGUESS,
    ('WFC3',  'UVIS'): WFC3_UVIS_DEFAULT_RECTSIZEGUESS,
    ('WFC3',  'IR'):   WFC3_IR_DEFAULT_RECTSIZEGUESS,
}

DEFAULT_FOV = {
    ('ACS',   'HRC'):  ACS_HRC_DEFAULT_FOV,
    ('WFPC2', 'PC1'):  WFPC2_PC1_DEFAULT_FOV,
    ('WFC3',  'UVIS'): WFC3_UVIS_DEFAULT_FOV,
    ('WFC3',  'IR'):   WFC3_IR_DEFAULT_FOV,
}

DEFAULT_JITTER = {
    ('ACS',   'HRC'):  ACS_HRC_DEFAULT_JITTER,
    ('WFPC2', 'PC1'):  WFPC2_PC1_DEFAULT_JITTER,
    ('WFC3',  'UVIS'): WFC3_UVIS_DEFAULT_JITTER,
    ('WFC3',  'IR'):   WFC3_IR_DEFAULT_JITTER,
}

DEFAULT_SUBSAMPLE = {
    ('ACS',   'HRC'):  ACS_HRC_DEFAULT_SUBSAMPLE,
    ('WFPC2', 'PC1'):  WFPC2_PC1_DEFAULT_SUBSAMPLE,
    ('WFC3',  'UVIS'): WFC3_UVIS_DEFAULT_SUBSAMPLE,
    ('WFC3',  'IR'):   WFC3_IR_DEFAULT_SUBSAMPLE,
}

DEFAULT_LINE = {
    ('ACS',   'HRC'):  ACS_HRC_DEFAULT_LINE,
    ('WFPC2', 'PC1'):  WFPC2_PC1_DEFAULT_LINE,
    ('WFC3',  'UVIS'): WFC3_UVIS_DEFAULT_LINE,
    ('WFC3',  'IR'):   WFC3_IR_DEFAULT_LINE,
}

DEFAULT_SAMPLE = {
    ('ACS',   'HRC'):  ACS_HRC_DEFAULT_SAMPLE,
    ('WFPC2', 'PC1'):  WFPC2_PC1_DEFAULT_SAMPLE,
    ('WFC3',  'UVIS'): WFC3_UVIS_DEFAULT_SAMPLE,
    ('WFC3',  'IR'):   WFC3_IR_DEFAULT_SAMPLE,
}


#===============================================================================
#
# ENVIRONMENT VARIABLES AND UTILITY ROUTINES
#
#===============================================================================

TINY_TIM_DIR = os.environ['TINYTIM']
PSF_CACHE_DIR = os.environ['PSF_CACHE_DIR']

if os.getcwd()[1] == ':':
    # Windows
    DEV_NULL = 'NUL'
else:
    # Linux
    DEV_NULL = '/dev/null'

def path_join(*args):
    """A prettier version of os.path.join that works better under Windows."""
    return os.path.join(*args).replace('\\', '/')


#===============================================================================
#
# THE HSTPSF CLASS
#
#===============================================================================

class HSTPSF(PSF):
    """A PSF that represents one of the HST cameras."""

    def __init__(self, instrument, detector, filter, line=None, sample=None,
                 rect_size_guess=None, fov=None, subsample=None, movement=None,
                 movement_granularity=0.3,
                 jitter_x=None, jitter_y=None, jitter_z=None, **kwargs):
        """Create an HSTPSF object.

        Input:
            instrument        The HST instrument (ACS, WFPC2, WFC3)

            detector          The HST detector
                                  For ACS: HRC
                                  For WFPC2: PC1
                                  For WFC3: UVIS, IR

            filter            The HST filter

            line              The line (Y) coordinate of the center of the PSF;
                              used for off-axis instruments because of
                              location-specific distortions

            sample            The sample (X) coordinate of the center of the
                              PSF; used for off-axis instruments because of
                              location-specific distortions

            rect_size_guess   A guess for the largest PSF size that's going
                              to be asked for later; used to cache the initial
                              PSF so that it doesn't need to be recomputed
                              multiple times

            fov               The field of view of the camera used to generate
                              the PSF from TinyTim in arcsec;
                              recommended max 3.0

            subsample         Override the default subsample amount; only valid
                              for ACS; Must be odd

            movement          Motion blur in the (Y,X) direction. Must be a
                              tuple of scalars.

            movement_granularity
                              The number of pixels to step for each smear doing
                              motion blur. The default of 0.1 is empirically a
                              good choice for speed vs. accuracy (<< 0.01
                              pixels).

            jitter_x/y/z      The amount of jitter (in mas) to apply to the
                              PSF; z is the angle; None means use the default
                              for the instrument/detector
        """

        PSF.__init__(self, movement, movement_granularity)

        self.instrument = instrument
        self.detector = detector
        self.filter = filter
        self.line = line
        self.sample = sample
        self.rect_size_guess = rect_size_guess
        self.fov = fov
        self.subsample = subsample
        self.jitter_x = jitter_x
        self.jitter_y = jitter_y
        self.jitter_z = jitter_z

        self.do_subsample_diffusion = False # Only for certain instruments

        if self.line is None:
            self.line = DEFAULT_LINE[(self.instrument, self.detector)]
        if self.sample is None:
            self.sample = DEFAULT_SAMPLE[(self.instrument, self.detector)]
        if self.rect_size_guess is None:
            self.rect_size_guess = DEFAULT_RECTSIZEGUESS[(self.instrument,
                                                          self.detector)]
        if self.fov is None:
            self.fov = DEFAULT_FOV[(self.instrument, self.detector)]
        if self.subsample is None:
            self.subsample = DEFAULT_SUBSAMPLE[(self.instrument,
                                                self.detector)]
        assert self.subsample % 2 == 1 # Must be odd
        if self.jitter_x is None:
            self.jitter_x = DEFAULT_JITTER[(self.instrument, self.detector)]
        if self.jitter_y is None:
            self.jitter_y = DEFAULT_JITTER[(self.instrument, self.detector)]
        if self.jitter_z is None:
            self.jitter_z = 0.

        self.cached_psf_size = None
        self.cached_psf = None
        self.cached_offset = None
        self.cached_spline_order = None
        self.cached_pixelated_psf = None

        if instrument == 'ACS':
            self._init_ACS(**kwargs)
        elif instrument == 'WFPC2':
            self._init_WFPC2(**kwargs)
        elif instrument == 'WFC3':
            self._init_WFC3(**kwargs)
        else:
            print('UNKNOWN INSTRUMENT', instrument)
            assert False

    def _init_ACS(self, **kwargs):
        assert self.detector == 'HRC'
        self.do_subsample_diffusion = True

    def _init_WFPC2(self, **kwargs):
        assert self.detector == 'PC1'
        self.do_subsample_diffusion = False

    def _init_WFC3(self, **kwargs):
        assert self.detector == 'UVIS' or self.detector == 'IR'
        self.do_subsample_diffusion = True

        # See http://www.stsci.edu/hst/observatory/apertures/wfc3.html
        self.aperture = kwargs['aperture']

        if self.detector == 'UVIS':
            if self.aperture == 'UVIS2-C512C-SUB':
                x_offset = 0  # Lower left corner of chip 2 is at (0,0)
                y_offset = 0 +2048 # Will be taken off in run_tinytim after
                                   # chip number is detected
                self.wfc3_uvis_chip = 2
                if self.line is None:
                    self.line = 256
                if self.sample is None:
                    self.sample = 256
            elif self.aperture == 'UVIS1-C512B-SUB':
                x_offset = 4096-512
                y_offset = 2048-512  # Upper right corner of chip 1 is at
                                     # (4095, 2047)
                self.wfc3_uvis_chip = 1
                if self.line is None:
                    self.line = 256
                if self.sample is None:
                    self.sample = 256
            elif self.aperture == 'UVIS2-M512C-SUB':
                x_offset = 2048-512  # Lower left corner of chip 2 is at (0,0)
                y_offset = 2048-512 +2048 # Will be taken off in run_tinytim
                                          # after chip number is detected
                self.wfc3_uvis_chip = 2
                if self.line is None:
                    self.line = 256
                if self.sample is None:
                    self.sample = 256
            else:
                print('UNKNOWN WFC3 UVIS APERTURE', self.aperture)
                assert False
        else:
            if self.aperture != 'IRSUB256':
                print('UNKNOWN WFC3 IR APERTURE', self.aperture)
                assert False
            x_offset = 1014//2-256//2
            y_offset = 1014//2-256//2
            if self.line is None:
                self.line = 128
            if self.sample is None:
                self.sample = 128

        self.line += y_offset
        self.sample += x_offset

    @staticmethod
    def run_tinytim(instrument, detector, y_ctr, x_ctr, filter,
                    psf_size_pixels, fov, subsample_amt,
                    jitter_x=0., jitter_y=0., jitter_z=0.,
                    force_run_tinytim=False, output_dev_null=True,
                    **kwargs):
        """Run TinyTim to produce a PSF.

        Note that the TinyTim PSF is centered on the center pixel, not on the
        corner.

        Input:

        instrument         The HST instrument (ACS, WFC3)

        detector           The HST detector
                               ACS: HRC
                               WFPC2: PC1
                               WFC3: UVIS, IR

        y_ctr              The center of the PSF in the Y dimension on the
                           image plane

        x_ctr              The center of the PSF in the X dimension on the
                           image plane

        filter             The filter name (case irrelevant)

        psf_size_pixels    The desired size of the returned PSF in pixels
                           (the PSF is square)

        fov                The camera FOV in arcsec used to create the PSF

        subsample_amt      The amount of subsampling desired (1-10)
                              [ACS and WFC3 ONLY!]
                           NOTE: Subsampling > 1 means the charge diffusion
                                 has not been done!

        jitter_x/y/z       The amount of jitter (in mas) to apply to the PSF;
                           z is the angle

        force_run_tinytim  If True, rerun TinyTim even if there is a cached
                           version.

        Return:

        The PSF as an array of size (psf_size_pixels, psf_size_pixels)
        The convolution matrix required for charge diffusion if
        subsample_amt>1, or None
        """

        if instrument == 'ACS':
            assert detector == 'HRC'
        elif instrument == 'WFPC2':
            assert detector == 'PC1'
        elif instrument == 'WFC3':
            assert detector == 'UVIS' or detector == 'IR'
        else:
            assert False

        min_fov = None
        for fov_size, pixel_size in RETURNED_SIZES[(instrument, detector)]:
            if pixel_size > psf_size_pixels:
                min_fov = fov_size
                break
        assert min_fov is not None

        if min_fov*.7 > fov: # Give a little slack for the distortion
            print('WARNING: OVERRIDING SPECIFIED FOV', fov, end=' ')
            print('BECAUSE REQUESTED PSF IS TOO BIG')
            print('NEW FOV', min_fov, 'PSF SIZE', psf_size_pixels)
            fov = min_fov

        x_ctr = int(x_ctr)
        y_ctr = int(y_ctr)
        filter = filter.upper()

        if jitter_x != 0. and jitter_y != 0.:
            fits_base = ('%s_%s_%s_%04d_%04d_%04.2f_%02d_%06.2f_%06.2f_'
                         '%06.2f_00.fits' % (instrument, detector, filter,
                                             y_ctr, x_ctr,
                                             fov, subsample_amt,
                                             jitter_x, jitter_y, jitter_z))
        else:
            fits_base = ('%s_%s_%s_%04d_%04d_%4.2f_%02d_00.fits' %
                             (instrument, detector, filter, y_ctr, x_ctr,
                              fov, subsample_amt))
        fits_filename = path_join(PSF_CACHE_DIR, fits_base)
        if not os.path.exists(fits_filename) or force_run_tinytim:
            assert psf_size_pixels % 2 == 1 # Must be odd
            orig_cwd = os.getcwd()
            os.chdir(TINY_TIM_DIR)
            psf_filename = 'temp_psf' + str(os.getpid()) + '_'
            full_psf_filename = psf_filename + '00_psf.fits'
            temp_fits_filename = psf_filename + '00.fits'
            params_filename = 'params_input' + str(os.getpid()) + '.txt'
            temp_filename = 'temp' + str(os.getpid()) + '.in'

            params_fp = open(params_filename, 'w')

            if instrument == 'ACS' and detector == 'HRC':
                print(16, file=params_fp)               # ACS HRC
            elif instrument == 'WFPC2' and detector == 'PC1':
                print(6, file=params_fp)                # WFPC2 PC1
            elif instrument == 'WFC3' and detector == 'UVIS':
                print(22, file=params_fp)               # WFC3 UVIS
                if y_ctr > 2047:
                    print(2, file=params_fp)            # UVIS chip (1 or 2)
                    y_ctr -= 2048
                else:
                    print(1, file=params_fp)            # UVIS chip (1 or 2)
            elif instrument == 'WFC3' and detector == 'IR':
                print(23, file=params_fp)               # WFC3 IR

            print(x_ctr, y_ctr, file=params_fp)     # PSF location on detector
            if filter.lower() == 'fqch4n15' and instrument == 'WFPC2':
                # This filter is not in the tinytim database
                print('mono', file=params_fp)       # Monochromatic
                print(620, file=params_fp)          # nm
            elif filter.lower() == 'fqch4p15' and instrument == 'WFPC2':
                # This filter is not in the tinytim database
                print('mono', file=params_fp)       # Monochromatic
                print(893, file=params_fp)          # nm
            else:
                print(filter, file=params_fp)           # Filter
                if filter[-1].lower() != 'n':
                    # Only wide-band filters require an input spectrum
                    print(1, file=params_fp)            # Spectrum # (stars)
                    print(11, file=params_fp)           # G2V - our Sun
            print(fov, file=params_fp)  # PSF size in arcsec
            if instrument == 'WFPC2':
                if subsample_amt == 1:
                    print('N', file=params_fp)          # Subsample NO
                else:
                    print('Y', file=params_fp)          # Subsample YES
                    print(subsample_amt, file=params_fp)
            print(0, file=params_fp)                # Focus offset
            print(psf_filename, file=params_fp)     # PSF filename
            params_fp.close()

            if output_dev_null:
                redir = ' > ' + DEV_NULL
            else:
                print('*** PARAMS FOR TINY1 ***')
                os.system('cat '+params_filename)
                redir = ''
            if jitter_x != 0. and jitter_y != 0.:
                os.system('./tiny1 ' + temp_filename + ' major='+str(jitter_x) +
                          ' minor='+str(jitter_y) + ' angle='+str(jitter_z) +
                          ' < ' + params_filename + redir)
            else:
                os.system('./tiny1 ' + temp_filename + ' < ' + params_filename +
                          ' > ' + DEV_NULL)

            os.system('./tiny2 ' + temp_filename + redir)

            if instrument == 'ACS' or instrument == 'WFC3':
                if not output_dev_null:
                    print('*** TINY3 SUB='+str(subsample_amt)+' ***')
                os.system('./tiny3 ' + temp_filename + ' SUB=' +
                          str(subsample_amt) + redir)
                os.unlink(psf_filename + '.tt3')
                os.unlink(full_psf_filename)

            if not os.path.exists(temp_fits_filename):
                print('RUN_TINYTIM: ERROR CREATING FITS FILE', psf_filename)
                assert False

            psf_file = pyfits.open(temp_fits_filename)
            psf_file.writeto(fits_filename, overwrite=True)
            psf_file.close()

            os.unlink(temp_fits_filename)
            os.unlink(params_filename)
            os.unlink(temp_filename)

            os.chdir(orig_cwd)

        psf_file = pyfits.open(fits_filename)
        psf_data = psf_file[0].data
        psf_file.close()

        diffusion_matrix = None
        if subsample_amt > 1 and (instrument == 'ACS' or instrument == 'WFC3'):
            # Subsample > 1 requires later convolution for charge diffusion
            # We need to extract the convolution matrix from the header
            # comments
            comment_list = psf_file[0].header['COMMENT']
            diffusion_matrix = np.empty((3,3))
            for diffusion_y in range(3):
                comment_str = comment_list[diffusion_y+3].strip()
                comment_vals = comment_str.split(' ')
                for diffusion_x in range(3):
                    diffusion_matrix[diffusion_y, diffusion_x] = \
                                float(comment_vals[diffusion_x])

        psf_size = psf_data.shape
        psf_size_pixels *= subsample_amt
        psf_halfsize = psf_size_pixels//2
        if (psf_halfsize > psf_size[0]//2 or
            psf_halfsize > psf_size[1]//2):
            print('FATAL ERROR: TINYTIM returned a PSF smaller', psf_size, 'than we wanted', psf_halfsize)
            assert False
        psf_data = psf_data[psf_size[1]//2-psf_halfsize:
                            psf_size[1]//2+psf_halfsize+1,
                            psf_size[0]//2-psf_halfsize:
                            psf_size[0]//2+psf_halfsize+1]

        return psf_data, diffusion_matrix

    #===============================================================================
    # Instrument-independent private routines
    #===============================================================================

    def _cache_psf(self, min_size, **kwargs):
        max_xy = max(min_size, self.rect_size_guess)
        max_xy = (max_xy//2)*2+1 # Make it odd
        if self.cached_psf_size is None or max_xy > self.cached_psf_size:
            psf, diffusion_matrix = HSTPSF.run_tinytim(self.instrument,
                                                       self.detector,
                                                       self.line, self.sample,
                                                       self.filter,
                                                       max_xy, self.fov,
                                                       self.subsample,
                                                       self.jitter_x,
                                                       self.jitter_y,
                                                       self.jitter_z,
                                                       **kwargs)
            self.cached_psf_size = psf.shape[0]
            self.cached_psf = psf
            self.cached_diffusion_matrix = diffusion_matrix
            assert psf.shape[0] % 2 == 1 # Odd
            assert psf.shape[1] % 2 == 1 # Odd
            self.psf_zero_offset = psf.shape[0]//2
            self.cached_pixelated_psf = None

    def _cache_pixelation(self, offset, **kwargs):
        spline_order = 3
        if 'spline_order' in kwargs:
            spline_order = kwargs['spline_order']

        # We need (0,0) to be the upper left corner of the pixel
        offset = (offset[0]-0.5, offset[1]-0.5)

        if (self.cached_offset == offset and
            self.cached_spline_order == spline_order and
            self.cached_pixelated_psf is not None):
            return

        self.cached_offset = offset
        self.cached_spline_order = spline_order

        subsample_psf_size = self.cached_psf.shape[0]
        psf_size = subsample_psf_size // self.subsample
        # This is NOT the same as subsample_psf_size!
        psf_size_resampled = psf_size * self.subsample

        # First: Make a new subsampled PSF from the existing subsampled PSF
        # but offset by offset_y,offset_x using linear interpolation. This
        # includes doing motion blur.

        # The X and Y indices in the subsampled PSF
        native_indices = (np.arange(float(subsample_psf_size))-
                          self.psf_zero_offset) / self.subsample

        # Create a linear interpolation spline
        spline = RectBivariateSpline(native_indices, native_indices,
                                     self.cached_psf,
                                     kx=spline_order, ky=spline_order)

        num_steps = int(max(abs(self.movement[0])/self.movement_granularity,
                            abs(self.movement[1])/self.movement_granularity))

        if num_steps == 0:
            step_y = 0.
            step_x = 0.
        else:
            step_y = self.movement[0] / num_steps
            step_x = self.movement[1] / num_steps

        total_rect = None

        for step in range(num_steps+1):
            y = offset[0] + step_y*(step-num_steps/2.)
            x = offset[1] + step_x*(step-num_steps/2.)

            # The X and Y indices in our new offset coordinate system in the
            # subsampled PSF
            desired_y_indices = native_indices - y
            desired_x_indices = native_indices - x
            desired_y_indices = np.repeat(desired_y_indices,
                                          psf_size_resampled)
            desired_x_indices = np.tile(desired_x_indices,
                                        psf_size_resampled)

            interp_psf = spline.ev(desired_y_indices, desired_x_indices)
            interp_psf = interp_psf.reshape((psf_size_resampled,
                                             psf_size_resampled))

            if total_rect is None:
                total_rect = interp_psf
            else:
                total_rect += interp_psf

        total_rect /= float(num_steps+1)

        # Second: Downsample the offset subsampled PSF by taking the sum over
        # adjacent pixels
        psf = total_rect.reshape([psf_size, self.subsample, psf_size,
                                  self.subsample])
        psf = np.sum(psf, axis=3)
        psf = np.sum(psf, axis=1)

        if self.subsample > 1 and self.do_subsample_diffusion:
            # We have to do the charge diffusion ourselves
            psf = scisig.convolve2d(psf, self.cached_diffusion_matrix)
        self.cached_pixelated_psf = psf
        self.pixelated_psf_zero_offset = psf.shape[0] // 2


    #===============================================================================
    # Public routines
    #===============================================================================

    def eval_point(self, point, scale=1., base=0.):
        """Evaluate the PSF at a single, fractional, point.

        (0,0) is the center of the PSF and x and y may be negative.

        Input:
            coord       the coordinate (y,x) at which to evaluate the PSF.
            scale       a scale factor to apply to the resulting PSF.
            base        a scalar added to the resulting PSF.
        """

        assert False

    def eval_pixel(self, coord, offset=(0.,0.), scale=1., base=0., **kwargs):
        """Evaluate the PSF integrated over an entire integer pixel.

        The returned array has the PSF offset from the center by
        (offset_y,offset_x). An offset of (0,0) places the PSF in the upper
        left corner of the center pixel while an offset of (0.5,0.5)
        places the PSF in the center of the center pixel.

        NOTE: The pixel value includes the effect of motion blur.

        Input:
            coord       the integer coordinate (y,x) at which to evaluate the
                        PSF.
            offset      the amount (offset_y,offset_x) to offset the center
                        of the PSF.
            scale       a scale factor to apply to the resulting PSF.
            base        a scalar added to the resulting PSF.
        """

        self._cache_psf(max(abs(y)*2+1, abs(x)*2+1), **kwargs)
        self._cache_pixelation(offset, **kwargs)

        index_y = coord[0]+self.pixelated_psf_zero_offset
        index_x = coord[1]+self.pixelated_psf_zero_offset

        return self.cached_pixelated_psf[index_y, index_x] * scale + base

    def eval_rect(self, rect_size, offset=(0.,0.), scale=1., base=0., **kwargs):
        """Create a rectangular pixelated PSF.

        This is done by evaluating the PSF function from:
            [-rect_size_y//2:rect_size_y//2] and
            [-rect_size_x//2:rect_size_x//2]

        The returned array has the PSF offset from the center by
        (offset_y,offset_x). An offset of (0,0) places the PSF in the upper
        left corner of the center pixel while an offset of (0.5,0.5)
        places the PSF in the center of the center pixel.

        Input:
            rect_size   the size of the rectangle (rect_size_y,rect_size_x)
                        of the returned PSF. Both dimensions must be odd.
            offset      the amount (offset_y,offset_x) to offset the center
                        of the PSF from the upper left corner of the center
                        pixel of the rectangle.
            scale       a scale factor to apply to the resulting PSF.
            base        a scalar added to the resulting PSF.
        """

        rect_size_y, rect_size_x = rect_size

        assert rect_size_y % 2 == 1 # Odd
        assert rect_size_x % 2 == 1 # Odd

        half_rect_size_y = rect_size_y // 2
        half_rect_size_x = rect_size_x // 2

        self._cache_psf(max(rect_size_x, rect_size_y), **kwargs)
        self._cache_pixelation(offset, **kwargs)

        rect = self.cached_pixelated_psf[self.pixelated_psf_zero_offset-half_rect_size_y:
                                         self.pixelated_psf_zero_offset+half_rect_size_y+1,
                                         self.pixelated_psf_zero_offset-half_rect_size_x:
                                         self.pixelated_psf_zero_offset+half_rect_size_x+1]

        rect = np.maximum(rect, 0.)

        return rect * scale + base



'''
Created on Jan 5, 2012

@author: rfrench
'''

try:
    from tkinter import *
except ImportError:
    from Tkinter import *

import numpy as np
from psfmodel.gaussian import GaussianPSF
from psfmodel.hst import HSTPSF


display_size = 32 # 64

# PSF_TYPE = 'gaussian'
PSF_TYPE = 'acshrc'
# PSF_TYPE = 'wfc3uvis'
# PSF_TYPE = 'wfpc2pc1'

psfobj = None

def command_refresh_psf(val):
    global psfobj

    if (psfobj is None or
        (PSF_TYPE == 'gaussian' and
         (psfobj.sigma_x != var_sigmax.get() or
          psfobj.sigma_y != var_sigmay.get() or
          psfobj.movement != (var_motiony.get(), var_motionx.get()))) or
        ((PSF_TYPE in ('acshrc', 'wfpc2pc1', 'wfc3uvis') and
         (psfobj.subsample != var_subsample.get() or
          psfobj.movement != (var_motiony.get(), var_motionx.get()))))):
        if PSF_TYPE == 'gaussian':
            print((var_motiony.get(), var_motionx.get()))
            psfobj = GaussianPSF((var_sigmay.get(), var_sigmax.get()),
                                 movement=(var_motiony.get(),
                                           var_motionx.get()))
        elif PSF_TYPE == 'acshrc':
            psfobj = HSTPSF('ACS', 'HRC', 'F660N', 512, 512,
                            subsample=var_subsample.get()*2+1,
                            movement=(var_motiony.get(), var_motionx.get()))
        elif PSF_TYPE == 'wfc3uvis':
            psfobj = HSTPSF('WFC3', 'UVIS', 'F606W', 128, 128,
                            subsample=var_subsample.get()*2+1,
                            movement=(var_motiony.get(), var_motionx.get()),
                            aperture='UVIS2-C512C-SUB')
        elif PSF_TYPE == 'wfpc2pc1':
            psfobj = HSTPSF('WFPC2', 'PC1', 'F606W', 128, 128,
                            subsample=var_subsample.get()*2+1,
                            movement=(var_motiony.get(), var_motionx.get()),
                            aperture='UVIS2-C512C-SUB')

    psf = psfobj.eval_rect(((var_psf_ysize.get()//2)*2+1,
                            (var_psf_xsize.get()//2)*2+1),
                           (var_y.get(), var_x.get()),
                           output_dev_null=False)
    print('PSF SUM', np.sum(psf))
    psf = psf**.5 #np.log10(psf+1e-10)

#    bkgnd = astrofit.background_gradient((var_psf_xsize.get(), var_psf_ysize.get()),
#                                         var_bkgnd_bias.get(), var_bkgnd_scale.get(), var_bkgnd_angle.get())
#    psf += bkgnd

    pix_scale = canvas_size // display_size
    ctr_x = canvas_size // 2
    ctr_y = canvas_size // 2
    min_val = 0 # np.min(psf)
    max_val = np.max(psf)
    canvas.delete('rect')
    for y in range(psf.shape[0]):
        for x in range(psf.shape[1]):
            val = int(max((psf[y, x]-min_val) / (max_val-min_val) * 255, 0))
            color = '#%02x%02x%02x' % (val, val, val)
            canvas.create_rectangle(((x-psf.shape[1]//2)*pix_scale+ctr_x,
                                     (y-psf.shape[0]//2)*pix_scale+ctr_y,
                                     (x-psf.shape[1]//2+1)*pix_scale+ctr_x,
                                     (y-psf.shape[0]//2+1)*pix_scale+ctr_y),
                                    outline=color, fill=color, tags='rect')

if __name__ == "__main__":
    toplevel_frame = Frame()
    canvas_size = 512
    canvas = Canvas(toplevel_frame, width=canvas_size, height=canvas_size, bg='black',
                    cursor='crosshair')
    canvas.grid(row=0, column=0, sticky=NW)

    # Control sliders
    control_frame = Frame(toplevel_frame)

    var_x = DoubleVar()
    var_x.set(0.)
    var_y = DoubleVar()
    var_y.set(0.)
    var_sigmax = DoubleVar()
    var_sigmax.set(2.)
    var_sigmay = DoubleVar()
    var_sigmay.set(2.)
    var_psf_xsize = IntVar()
    var_psf_xsize.set(21)
    var_psf_ysize = IntVar()
    var_psf_ysize.set(21)
    var_subsample = IntVar()
    var_subsample.set(0)
    var_motionx = DoubleVar()
    var_motionx.set(0.)
    var_motiony = DoubleVar()
    var_motiony.set(0.)
    var_bkgnd_bias = DoubleVar()
    var_bkgnd_bias.set(0.)
    var_bkgnd_scale = DoubleVar()
    var_bkgnd_scale.set(0.)
    var_bkgnd_angle = DoubleVar()
    var_bkgnd_angle.set(0.)

    gridrow = 0

    label = Label(control_frame, text='X')
    label.grid(row=gridrow, column=0, sticky=W)
    scale_x = Scale(control_frame, orient=HORIZONTAL, from_=-5., to=5., resolution=0.01,
                    variable=var_x, command=command_refresh_psf)
    scale_x.grid(row=gridrow, column=1)
    gridrow += 1

    label = Label(control_frame, text='Y')
    label.grid(row=gridrow, column=0, sticky=W)
    scale_y = Scale(control_frame, orient=HORIZONTAL, from_=-5., to=5., resolution=0.01,
                    variable=var_y, command=command_refresh_psf)
    scale_y.grid(row=gridrow, column=1)
    gridrow += 1

    if PSF_TYPE == 'gaussian':
        label = Label(control_frame, text='SIGMA X')
        label.grid(row=gridrow, column=0, sticky=W)
        scale_sigmax = Scale(control_frame, orient=HORIZONTAL, from_=0.001, to=5., resolution=0.001,
                            variable=var_sigmax, command=command_refresh_psf)
        scale_sigmax.grid(row=gridrow, column=1)
        gridrow += 1

        label = Label(control_frame, text='SIGMA Y')
        label.grid(row=gridrow, column=0, sticky=W)
        scale_sigmay = Scale(control_frame, orient=HORIZONTAL, from_=0.001, to=5., resolution=0.001,
                            variable=var_sigmay, command=command_refresh_psf)
        scale_sigmay.grid(row=gridrow, column=1)
        gridrow += 1

    # label = Label(control_frame, text='BKGND BIAS')
    # label.grid(row=gridrow, column=0, sticky=W)
    # scale_bkgnd_bias = Scale(control_frame, orient=HORIZONTAL, from_=0., to=1, resolution=0.001,
    #                      variable=var_bkgnd_bias, command=command_refresh_psf)
    # scale_bkgnd_bias.grid(row=gridrow, column=1)
    # gridrow += 1
    #
    # label = Label(control_frame, text='BKGND SCALE')
    # label.grid(row=gridrow, column=0, sticky=W)
    # scale_bkgnd_scale = Scale(control_frame, orient=HORIZONTAL, from_=0., to=.05, resolution=0.001,
    #                           variable=var_bkgnd_scale, command=command_refresh_psf)
    # scale_bkgnd_scale.grid(row=gridrow, column=1)
    # gridrow += 1
    #
    # label = Label(control_frame, text='BKGND ANGLE')
    # label.grid(row=gridrow, column=0, sticky=W)
    # scale_bkgnd_angle = Scale(control_frame, orient=HORIZONTAL, from_=0., to=360., resolution=1.,
    #                           variable=var_bkgnd_angle, command=command_refresh_psf)
    # scale_bkgnd_angle.grid(row=gridrow, column=1)
    # gridrow += 1

    gridrow = 0

    label = Label(control_frame, text='PSF X SIZE')
    label.grid(row=gridrow, column=2, sticky=W)
    scale_psf_xsize = Scale(control_frame, orient=HORIZONTAL, from_=1, to=101., resolution=1,
                            variable=var_psf_xsize, command=command_refresh_psf)
    scale_psf_xsize.grid(row=gridrow, column=3)
    gridrow += 1

    label = Label(control_frame, text='PSF Y SIZE')
    label.grid(row=gridrow, column=2, sticky=W)
    scale_psf_ysize = Scale(control_frame, orient=HORIZONTAL, from_=1, to=101., resolution=1,
                            variable=var_psf_ysize, command=command_refresh_psf)
    scale_psf_ysize.grid(row=gridrow, column=3)
    gridrow += 1

    label = Label(control_frame, text='SUBSAMPLE (*2+1)')
    label.grid(row=gridrow, column=2, sticky=W)
    scale_subsample = Scale(control_frame, orient=HORIZONTAL, from_=0, to=4., resolution=1,
                            variable=var_subsample, command=command_refresh_psf)
    scale_subsample.grid(row=gridrow, column=3)
    gridrow += 1

    label = Label(control_frame, text='MOTION X')
    label.grid(row=gridrow, column=2, sticky=W)
    scale_motionx = Scale(control_frame, orient=HORIZONTAL, from_=-10., to=10., resolution=.1,
                        variable=var_motionx, command=command_refresh_psf)
    scale_motionx.grid(row=gridrow, column=3)
    gridrow += 1

    label = Label(control_frame, text='MOTION Y')
    label.grid(row=gridrow, column=2, sticky=W)
    scale_motiony = Scale(control_frame, orient=HORIZONTAL, from_=-10., to=10., resolution=.1,
                        variable=var_motiony, command=command_refresh_psf)
    scale_motiony.grid(row=gridrow, column=3)
    gridrow += 1

    control_frame.grid(row=1, column=0, sticky=NW)
    toplevel_frame.pack()

    command_refresh_psf(0)

    mainloop()



################################################################################
# psfmodel/__init__.py
################################################################################

import numpy as np
import numpy.ma as ma
import scipy.linalg as linalg
import scipy.optimize as sciopt

try:
    from _version import __version__
except ImportError as err:
    __version__ = 'Version unspecified'


class PSF(object):
    """Abstract superclass for classes that model different types of PSFs."""

    def __init__(self, movement, movement_granularity):
        """Create a PSF object. Only called by subclasses.

        Input:
            movement          Motion blur in the (Y,X) direction. Must be a
                              tuple of scalars or None.
            movement_granularity
                              The number of pixels to step for each smear doing
                              motion blur.
        """

        self.debug_opt = 0

        if movement is None:
            movement = (0., 0.)
        self.movement = movement
        self.movement_granularity = movement_granularity

        self.additional_params = []

    def eval_point(self, coord, scale=1., base=0., **kwargs):
        """Evaluate the PSF at a single, fractional, point.

        (0,0) is the center of the PSF and x and y may be negative.

        Input:
            coord       The coordinate (y,x) at which to evaluate the PSF.

            scale       A scale factor to apply to the resulting PSF.
            base        A scalar added to the resulting PSF.

        Other inputs may be available for specific subclasses.
        """
        assert False

    def eval_pixel(self, coord, offset=(0.,0.), scale=1., base=0., **kwargs):
        """Evaluate the PSF integrated over an entire integer pixel.

        The returned array has the PSF offset from the center by
        (offset_y,offset_x). An offset of (0,0) places the PSF in the upper
        left corner of the center pixel while an offset of (0.5,0.5)
        places the PSF in the center of the center pixel.

        Input:
            coord       The integer coordinate (y,x) at which to evaluate the
                        PSF.
            offset      The amount (offset_y,offset_x) to offset the center
                        of the PSF.

            scale       A scale factor to apply to the resulting PSF.
            base        A scalar added to the resulting PSF.

        Other inputs may be available for specific subclasses.
        """
        assert False

    def eval_rect(self, rect_size, offset=(0.,0.), scale=1., base=0.,
                  **kwargs):
        """Create a rectangular pixelated PSF.

        This is done by evaluating the PSF function from:
            [-rect_size_y//2:rect_size_y//2] and
            [-rect_size_x//2:rect_size_x//2]

        The returned array has the PSF offset from the center by
        (offset_y,offset_x). An offset of (0,0) places the PSF in the upper
        left corner of the center pixel while an offset of (0.5,0.5)
        places the PSF in the center of the center pixel.

        Input:
            rect_size   The size of the rectangle (rect_size_y,rect_size_x)
                        of the returned PSF. Both dimensions must be odd.
            offset      The amount (offset_y,offset_x) to offset the center
                        of the PSF from the upper left corner of the center
                        pixel of the rectangle.

            scale       A scale factor to apply to the resulting PSF.
            base        A scalar added to the resulting PSF.

        Other inputs may be available for specific subclasses.
        """
        assert False

    def _eval_rect_smeared(self, movement, rect_size, offset=(0.,0.), *args,
                           **kwargs):
        """Evaluate and sum a PSF multiple times to simulate motion blur.

        Input:
            movement    The total amount (my,mx) the PSF moves. The movement
                        is assumed to be centered on the given offset and
                        exists half on either side.
            rect_size   The size of the rectangle (rect_size_y,rect_size_x)
                        of the returned PSF. Both dimensions must be odd.

            offset      The amount (offset_y,offset_x) to offset the center
                        of the PSF. A positive offset effectively moves the
                        PSF down and to the left.
        """

        num_steps = int(max(abs(movement[0])/self.movement_granularity,
                            abs(movement[1])/self.movement_granularity))

        if num_steps == 0:
            step_y = 0.
            step_x = 0.
        else:
            step_y = movement[0] / num_steps
            step_x = movement[1] / num_steps

        total_rect = None

        for step in range(num_steps+1):
            y = offset[0] + step_y*(step-num_steps/2.)
            x = offset[1] + step_x*(step-num_steps/2.)

            rect = self._eval_rect(rect_size, (y, x), *args, **kwargs)
            if total_rect is None:
                total_rect = rect
            else:
                total_rect += rect

        total_rect /= float(num_steps+1)

        return total_rect


    #==========================================================================
    #
    # Static functions for creating background gradients
    #
    #==========================================================================

    @staticmethod
    def _background_gradient_coeffs(shape, order):
        """Internal routine for creating the coefficient matrix."""

        lvalues = np.arange(shape[0])[:,np.newaxis]-int(shape[0]/2)
        svalues = np.arange(shape[1])[np.newaxis,:]-int(shape[1]/2)

        lpowers = [1.]
        spowers = [1.]

        nparams = int((order+1) * (order+2) / 2)
        a3d = np.empty((shape[0], shape[1], nparams))
        a3d[:,:,0] = 1.

        k = 0
        for p in range(1,order+1):
            lpowers.append(lpowers[-1] * lvalues)
            spowers.append(spowers[-1] * svalues)

            for q in range(p+1):
                k += 1
                a3d[:,:,k] = lpowers[q] * spowers[p-q]

        return a3d

    @staticmethod
    def background_gradient_fit(image, order=2, ignore_center=None,
                                num_sigma=None, debug=0):
        """Return the polynomial fit to the pixels of an image.

        Input:
            image           2D array to fit; must have odd shape in each
                            dimension.
            order           Order of the polynomial; default=2 (quadratic).
            ignore_center   A scalar or tuple (ignore_y,ignore_x) giving the
                            number of pixels on either side of the center to
                            ignore while fitting. 0 means ignore the center
                            pixel. None means don't ignore anything.
            num_sigma       The number of sigma a pixel needs to be beyond the
                            background gradient to be ignored. None means don't
                            ignore bad pixels.
            debug           Set to debug bad pixel removal.
        """

        assert image.shape[0] % 2 == 1 # Image shape must be odd
        assert image.shape[1] % 2 == 1

        is_masked = False

        if ignore_center or num_sigma:
            if isinstance(image, ma.MaskedArray):
                # We're going to change the mask so make a copy first
                image = image.copy()
            else:
                image = image.view(ma.MaskedArray)

        if isinstance(image, ma.MaskedArray):
            image.mask = ma.getmaskarray(image)
            is_masked = True

        if ignore_center:
            ctr_y = image.shape[0] // 2
            ctr_x = image.shape[1] // 2
            if np.shape(ignore_center) == ():
                ignore_y = ignore_center
                ignore_x = ignore_center
            else:
                ignore_y, ignore_x = ignore_center
            if (ignore_y*2+1 >= image.shape[0] or
                ignore_x*2+1 >= image.shape[1]):
                if debug:
                    print('BKGND CENTER IGNORED IS ENTIRE IMAGE')
                return None, None
            image[ctr_y-ignore_y:ctr_y+ignore_y+1,
                  ctr_x-ignore_x:ctr_x+ignore_x+1] = ma.masked

        nparams = int((order+1) * (order+2) // 2)

        a3d = PSF._background_gradient_coeffs(image.shape, order)

        if num_sigma:
            num_bad_pixels = ma.count_masked(image)
            if debug:
                print('BKGND GRAD INIT # BAD', num_bad_pixels)

        while True:
            # Reshape properly for linalg.lstsq
            a2d  = a3d.reshape((image.size, nparams))
            b1d  = image.flatten()

            if is_masked:
                a2d = a2d[~b1d.mask] # linalg doesn't support masked arrays!
                b1d = ma.compressed(b1d)

            if a2d.shape[0] < a2d.shape[1]: # Underconstrained
                if debug:
                    print('BKGND UNDERCONSTRAINED', a2d.shape)
                return None, None

            coeffts = linalg.lstsq(a2d, b1d)[0]

            if not num_sigma:
                break

            gradient = PSF.background_gradient(image.shape, coeffts)
            new_img = image-gradient
            sigma = np.std(new_img)
            image[np.where(np.abs(new_img) > sigma*num_sigma)] = ma.masked

            new_num_bad_pixels = ma.count_masked(image)
            if debug:
                print('BKGD GRAD NEW # BAD', new_num_bad_pixels)
            if new_num_bad_pixels == num_bad_pixels:
                break
            num_bad_pixels = new_num_bad_pixels

        if is_masked:
            return coeffts, ma.getmaskarray(image)
        else:
            return coeffts, np.zeros(image.shape)

    @staticmethod
    def background_gradient(size, bkgnd_params):
        """Create a background gradient.

        Input:
            size            A scalar (meaning a square) or a tuple
                            (size_y, size_x) indicating the size of the
                            returned array.
            bkgnd_params    A tuple indicating the coefficients of the
                            background polynomial. The order of the polynomial
                            is inferred from the number of elements in the
                            tuple.
        """

        if np.shape(size) == ():
            size_y = size
            size_x = size
        else:
            size_y, size_x = size

        bkgnd_params = np.array(bkgnd_params)

        order = int(np.sqrt(len(bkgnd_params)*2))-1

        a3d = PSF._background_gradient_coeffs(size, order)
        result = np.sum(bkgnd_params * a3d, axis=-1)

        return result

    #==========================================================================
    #
    # Functions for finding astrometric positions
    #
    #==========================================================================

    def find_position(self, image, box_size,
                      starting_point, search_limit=(1.5, 1.5),
                      bkgnd_degree=2, bkgnd_ignore_center=(2,2),
                      bkgnd_num_sigma=5,
                      tolerance=1e-8, num_sigma=10,
                      max_bad_frac=0.2,
                      allow_nonzero_base=False,
                      scale_limit=1000000.,
                      use_angular_params=True):
        """Find the (y,x) coordinates that best fit a 2-D Gaussian to an image.

        Input:
            image                The image (arbitrary 2-D size, floating
                                 point).

            box_size             A tuple (box_y,box_x) indicating the size of
                                 the PSF to use. This governs both the size of
                                 the PSF created at each step as well as the
                                 size of the subimage looked at. Both box_y
                                 and box_x must be odd.

            starting_point       A tuple (y,x) indicating the best guess for
                                 where the object can be found. Searching is
                                 limited to a region around this point
                                 controlled by "search_limit".

            search_limit         A scalar or tuple (y_limit,x_limit)
                                 specifying the maximum distance to search from
                                 "starting_point". If a scalar, both x_limit
                                 and y_limit are the same.

            bkgnd_degree         The degree (order) of the background gradient
                                 polynomial. None means no background gradient
                                 is fit.

            bkgnd_ignore_center  A tuple (y,x) giving the number of pixels on
                                 each side of the center point to ignore when
                                 fitting the background. The ignored region is
                                 thus n*2+1 on each side.

            bkgnd_num_sigma      The number of sigma a pixel needs to be beyond
                                 the background gradient to be ignored. None
                                 means don't ignore bad pixels while computing
                                 the background gradient.

            tolerance            The tolerance (both X and Function) in the
                                 Powell optimization algorithm.

            num_sigma            The number of sigma for a pixel to be
                                 considered bad during PSF fitting. None means
                                 don't ignore bad pixels while fitting the
                                 PSF.

            max_bad_frac         The maximum allowable number of pixels masked
                                 during PSF fitting. If more pixels than this
                                 fraction are masked, the position fit fails.

            allow_nonzero_base   If True, allow the base of the PSF (constant
                                 bias) to vary. Otherwise the base of the PSF
                                 is always at zero and can only scale in the
                                 positive direction.

            scale_limit          The maximum PSF scale allowed.

            use_angular_params   Use angles to optimize parameter values.

        Returns None if no fit found.

        Otherwise returns pos_y, pos_x, metadata. Metadata is a dictionary
        containing:

            'x'                    The offset in X. (Same as pos_x)
            'x_err'                Uncertainty in X.
            'y'                    The offset in Y. (Same as pos_y)
            'y_err'                Uncertainty in Y.
            'scale'                The best fit PSF scale.
            'scale_err'            Uncertainty in PSF scale.
            'base'                 The best fit PSF base.
            'base_err'             Uncertainty in PSF base.
            'subimg'               The box_size area of the original image
                                   surrounding starting_point masked as
                                   necessary using the num_sigma threshold.
            'bkgnd_params'         The tuple of parameters defining the
                                   background gradient.
            'bkgnd_mask'           The mask used during background gradient
                                   fitting.
            'gradient'             The box_size background gradient.
            'subimg-gradient'      The subimg with the background gradient
                                   subtracted.
            'psf'                  The unit-scale and zero-base PSF.
            'scaled_psf'           The fully scaled PSF with the base added.
            'leastsq_cov'          The covariance matrix returned by leastsq
                                   as adjusted by the residual variance.
            'leastsq_infodict'     The infodict returned by leastsq.
            'leastsq_mesg'         The mesg returned by leastsq.
            'leastsq_ier'          The ier returned by leastsq.

        In addition, metadata includes two entries for each "additional
        parameter" used during optimization: one for the value and one for
        the uncertainty ('param' and 'param_err').
        """

        assert box_size[0] % 2 == 1 # PSF box size must be odd
        assert box_size[1] % 2 == 1

        half_box_size_y = box_size[0] // 2
        half_box_size_x = box_size[1] // 2

        starting_pix = (int(np.round(starting_point[0])),
                        int(np.round(starting_point[1])))

        if self.debug_opt:
            print('>> Entering psfmodel:find_position')
            print('Image is masked', isinstance(image, ma.MaskedArray))
            print('Image num masked', np.sum(ma.getmaskarray(image)))
            print('Image min, max, mean', np.min(image), np.max(image), end=' ')
            print(np.mean(image))
            print('Box size', box_size)
            print('Starting point', starting_point)
            print('Search limit', search_limit)
            print('Bkgnd degree', bkgnd_degree)
            print('Bkgnd ignore center', bkgnd_ignore_center)
            print('Bkgnd num sigma', bkgnd_num_sigma)
            print('Tolerance', tolerance)
            print('Num sigma', num_sigma)
            print('Max bad frac', max_bad_frac)
            print('Allow nonzero base', allow_nonzero_base)
            print('Scale limit', scale_limit)
            print('Use angular params', use_angular_params)
            print('-----')

        # Too close to the edge means we can't search
        if (starting_pix[0]-half_box_size_y < 0 or
            starting_pix[0]+half_box_size_y >= image.shape[0] or
            starting_pix[1]-half_box_size_x < 0 or
            starting_pix[1]+half_box_size_x >= image.shape[1]):
            if self.debug_opt:
                print('Too close to the edge - search impossible')
            return None

        sub_img = image[starting_pix[0]-half_box_size_y:
                        starting_pix[0]+half_box_size_y+1,
                        starting_pix[1]-half_box_size_x:
                        starting_pix[1]+half_box_size_x+1]

        if self.debug_opt:
            print('Sub img min, max, mean', np.min(sub_img), end=' ')
            print(np.max(sub_img), np.mean(sub_img))

        if np.shape(search_limit) == ():
            search_limit = (float(search_limit), float(search_limit))

        if num_sigma:
            if isinstance(sub_img, ma.MaskedArray):
                # We're going to change the mask so make a copy first
                sub_img = sub_img.copy()
            else:
                sub_img = sub_img.view(ma.MaskedArray)

        num_bad_pixels = 0

        while True:
            if self.debug_opt > 1:
                print('MAIN LOOP: FIND POS # BAD PIXELS', num_bad_pixels)
            ret = self._find_position(sub_img,
                                      half_box_size_y, half_box_size_x,
                                      search_limit, scale_limit,
                                      bkgnd_degree, bkgnd_ignore_center,
                                      bkgnd_num_sigma, tolerance,
                                      allow_nonzero_base, use_angular_params)
            if ret is None:
                if self.debug_opt:
                    print('find_position returning None')
                return None

            res_y, res_x, details = ret

            if not num_sigma:
                break

            resid = np.sqrt((details['subimg-gradient']-
                             details['scaled_psf'])**2)
            resid_std = np.std(resid)

            if self.debug_opt > 1:
                print('MAIN LOOP: Resid', resid)
                print('resid_std', resid_std)

            if num_sigma is not None:
                sub_img[np.where(resid > num_sigma*resid_std)] = ma.masked

            new_num_bad_pixels = ma.count_masked(sub_img)
            if new_num_bad_pixels == num_bad_pixels:
                break
            if new_num_bad_pixels == sub_img.size:
                if self.debug_opt:
                    print('MAIN LOOP: All pixels masked - find_position returning None')
                return None # All masked
            if new_num_bad_pixels > max_bad_frac*sub_img.size:
                if self.debug_opt:
                    print('MAIN LOOP: Too many pixels masked - find_position returning None')
                return None # Too many masked
            num_bad_pixels = new_num_bad_pixels

        if self.debug_opt:
            print('find_position returning', end=' ')
            print('Y %.4f' % (res_y+starting_pix[0]), end=' ')
            if details['y_err'] is not None:
                print('+/- %.4f' % details['y_err'], end=' ')
            print('X %.4f' % (res_x+starting_pix[1]), end=' ')
            if details['x_err'] is not None:
                print('+/- %.4f' % details['x_err'], end=' ')
            print()
            if details['scale'] is not None:
                print('                        Scale %.4f Base %.4f' % (
                        details['scale'], details['base']))

        return res_y+starting_pix[0], res_x+starting_pix[1], details

    def _fit_psf_func(self, params, sub_img, search_limit, scale_limit,
                      allow_nonzero_base, use_angular_params,
                      *additional_params):
        if self.debug_opt > 1:
            print(params)
        # Make an offset of "0" be the center of the pixel (0.5,0.5)
        if use_angular_params:
            offset_y = search_limit[0] * np.sin(params[0]) + 0.5
            offset_x = search_limit[1] * np.sin(params[1]) + 0.5
            scale = scale_limit * (np.sin(params[2]) + 1) / 2
        else:
            offset_y = params[0] + 0.5
            offset_x = params[1] + 0.5
            scale = params[2]
            fake_resid = None
            if not -search_limit[0] <= params[0] <= search_limit[0]:
                fake_resid = abs(params[0]) * 1e10
            elif not -search_limit[1] <= params[1] <= search_limit[1]:
                fake_resid = abs(params[1]) * 1e10
            elif not 0.00001 <= scale <= scale_limit:
                fake_resid = abs(scale) * 1e10
            if fake_resid is not None:
                fake_return = np.zeros(sub_img.shape).flatten()
                fake_return[:] = fake_resid
                if self.debug_opt > 1:
                    full_resid = np.sqrt(np.sum(fake_return**2))
                    print('RESID', full_resid)
                return fake_return

        base = 0.
        param_end = 3
        if allow_nonzero_base:
            base = params[3]
            param_end = 4

        addl_vals_dict = {}
        for i, ap in enumerate(additional_params):
            if use_angular_params:
                val = ((ap[1] - ap[0]) / 2. *
                       (np.sin(params[param_end+i])+1.) + ap[0])
            else:
                val = params[param_end+i]
            addl_vals_dict[ap[2]] = val

        if self.debug_opt > 1:
            print(('OFFY %8.5f OFFX %8.5f SCALE %9.5f BASE %9.5f' %
                   (offset_y, offset_x, scale, base)), end=' ')
            for ap in additional_params:
                print('%s %8.5f' % (ap[2].upper(), addl_vals_dict[ap[2]]), end=' ')
            print()

        psf = self.eval_rect(sub_img.shape, (offset_y, offset_x),
                             scale=scale, base=base, **addl_vals_dict)

        resid = (sub_img-psf).flatten()

        if self.debug_opt > 1:
            full_resid = np.sqrt(np.sum(resid**2))
            print('RESID', full_resid)

        return resid

    def _find_position(self, sub_img, half_box_size_y, half_box_size_x,
                       search_limit, scale_limit, bkgnd_degree,
                       bkgnd_ignore_center, bkgnd_num_sigma, tolerance,
                       allow_nonzero_base, use_angular_params):

        bkgnd_params = None
        bkgnd_mask = None
        gradient = np.zeros(sub_img.shape)

        if bkgnd_degree is not None:
            bkgnd_params, bkgnd_mask = PSF.background_gradient_fit(
                                           sub_img,
                                           order=bkgnd_degree,
                                           ignore_center=bkgnd_ignore_center,
                                           num_sigma=bkgnd_num_sigma,
                                           debug=self.debug_opt > 1)
            if bkgnd_params is None:
                return None

            gradient = PSF.background_gradient(sub_img.shape, bkgnd_params)

        sub_img_grad = sub_img - gradient

        # Offset Y, Offset X, Scale, AdditionalParams
        if use_angular_params:
            starting_guess = [0.001,0.001,np.pi/8]
            if allow_nonzero_base:
                starting_guess = starting_guess + [0.001]
            starting_guess = starting_guess + [0.001] * len(self.additional_params)
        else:
            starting_guess = [0.001,0.001,1.]
            if allow_nonzero_base:
                starting_guess = starting_guess + [0.001]
            for a_min, a_max, a_name in self.additional_params:
                starting_guess = starting_guess + [np.mean([a_min,a_max])]

        extra_args = (sub_img_grad, search_limit, scale_limit,
                      allow_nonzero_base, use_angular_params)
        if (self.additional_params is not None and
            len(self.additional_params) > 0):
            extra_args = extra_args + tuple(self.additional_params)
        else:
            extra_args = extra_args + tuple([])
        full_result = sciopt.leastsq(self._fit_psf_func, starting_guess,
                                     args=extra_args,
                                     full_output=True,
                                     xtol=tolerance)

        result, cov_x, infodict, mesg, ier = full_result

        if ier < 1 or ier > 4:
            return None

        if use_angular_params:
            offset_y = search_limit[0] * np.sin(result[0]) + 0.5
            offset_x = search_limit[1] * np.sin(result[1]) + 0.5
            scale = scale_limit * (np.sin(result[2]) + 1) / 2
        else:
            offset_y = result[0] + 0.5
            offset_x = result[1] + 0.5
            scale = result[2]

        base = 0.
        result_end = 3
        if allow_nonzero_base:
            base = result[3]
            result_end = 4

        addl_vals_dict = {}
        for i, ap in enumerate(self.additional_params):
            if use_angular_params:
                val = ((ap[1] - ap[0]) / 2. *
                       (np.sin(result[result_end+i])+1.) + ap[0])
            else:
                val = result[result_end+i]
            addl_vals_dict[ap[2]] = val

        psf = self.eval_rect(sub_img.shape, (offset_y, offset_x),
                             **addl_vals_dict)

        details = {}
        details['x'] = offset_x
        details['y'] = offset_y
        details['subimg'] = sub_img
        details['bkgnd_params'] = bkgnd_params
        details['bkgnd_mask'] = bkgnd_mask
        details['gradient'] = gradient
        details['subimg-gradient'] = sub_img_grad
        details['psf'] = psf
        details['scale'] = scale
        details['base'] = base
        details['scaled_psf'] = psf*scale+base

        if cov_x is None:
            details['leastsq_cov'] = None
            details['x_err'] = None
            details['y_err'] = None
            details['scale_err'] = None
            details['base_err'] = None
            for i, ap in enumerate(self.additional_params):
                details[ap[2]+'_err'] = None
        else:
            # "To obtain the covariance matrix of the parameters x, cov_x must
            #  be multiplied by the variance of the residuals"
            dof = psf.shape[0]*psf.shape[1]-len(result)
            resid_var = np.sum(self._fit_psf_func(result, *extra_args)**2) / dof
            cov = cov_x * resid_var # In angle-parameter space!! (if use_angular_params)
            details['leastsq_cov'] = cov
            if use_angular_params:
                # Deriv of SL0 * sin(R0) is
                #   SL0 * cos(R0) * dR0
                y_err = np.abs((np.sqrt(cov[0][0]) % (np.pi*2)) * search_limit[0] * np.cos(result[0]))
                x_err = np.abs((np.sqrt(cov[1][1]) % (np.pi*2)) * search_limit[1] * np.cos(result[1]))
                # Deriv of SC/2 * (sin(R)+1) = SC/2 * sin(R) + SC/2 is
                #   SC/2 * cos(R) * dR
                scale_err = np.abs((np.sqrt(cov[2][2]) % (np.pi*2)) * scale_limit/2 * np.cos(result[2]))
                details['x_err'] = x_err
                details['y_err'] = y_err
                details['scale_err'] = scale_err
                for i, ap in enumerate(self.additional_params):
                    err = np.abs((np.sqrt(cov[i+3][i+3]) % (np.pi*2)) *
                                 (ap[1]-ap[0])/2 * np.cos(result[i+3]))
                    details[ap[2]+'_err'] = err
            else:
                details['x_err'] = np.sqrt(cov[0][0])
                details['y_err'] = np.sqrt(cov[1][1])
                details['scale_err'] = np.sqrt(cov[2][2])
                for i, ap in enumerate(self.additional_params):
                    details[ap[2]+'_err'] = np.sqrt(cov[i+result_end][i+result_end])
            # Note the base is not computed using angles
            details['base_err'] = None
            if allow_nonzero_base:
                details['base_err'] = np.sqrt(cov[3][3])

        details['leastsq_infodict'] = infodict
        details['leastsq_mesg'] = mesg
        details['leastsq_ier'] = ier

        for key in addl_vals_dict:
            details[key] = addl_vals_dict[key]

        if self.debug_opt > 1:
            print('_find_position RETURNING', offset_y, offset_x)
            print('Subimg num bad pixels', np.sum(ma.getmaskarray(sub_img)))
            print('Bkgnd params', bkgnd_params)
            print('Bkgnd mask bad pixels', np.sum(ma.getmaskarray(bkgnd_mask)))
            print('PSF scale', scale)
            print('PSF base', base)
            for key in addl_vals_dict:
                print(key, details[key])
            print('LEASTSQ COV')
            cov = details['leastsq_cov']
            print(cov)
            if cov is not None:
                print('X_ERR', details['x_err'])
                print('Y_ERR', details['y_err'])
                print('SCALE_ERR', details['scale_err'])
                print('BASE_ERR', details['base_err'])
                for key in addl_vals_dict:
                    print(key+'_err', details[key+'_err'])
            print('LEASTSQ MESG', mesg)
            print('LEASTSQ IER', ier)
            print('-----')

        return (offset_y, offset_x, details)



################################################################################
# UNIT TESTS
################################################################################

import unittest

import numpy as np
import numpy.ma as ma

from psfmodel import PSF


class Test_PSF(unittest.TestCase):

    def runTest(self):

        #
        # Background gradient fit
        #
        img = np.zeros((5,5))
        img[:] = 3*(np.arange(5.)-2)**2+2*(np.arange(5.)-2)
        bkgnd_params, img_mask = PSF.background_gradient_fit(img)
        self.assertTrue(np.sum((np.array(bkgnd_params)-
                                np.array([0,2,0,3,0,0]))**2) < 1e-10)
        self.assertEqual(np.sum(img_mask), 0)
        img2 = PSF.background_gradient((5,5), bkgnd_params)
        self.assertTrue(np.sum((img-img2)**2) < 1e-10)

        bkgnd_params, img_mask = PSF.background_gradient_fit(img, order=3)
        self.assertTrue(np.sum((np.array(bkgnd_params)-
                                np.array([0,2,0,3,0,0,0,0,0,0]))**2) < 1e-10)
        self.assertEqual(np.sum(img_mask), 0)
        img2 = PSF.background_gradient((5,5), bkgnd_params)
        self.assertTrue(np.sum((img-img2)**2) < 1e-10)

        # All masked
        img = np.zeros((5,5)).view(ma.MaskedArray)
        img[:,:] = ma.masked
        bkgnd_params, img_mask = PSF.background_gradient_fit(img)
        self.assertIsNone(bkgnd_params)
        self.assertIsNone(img_mask)

        # Ignore center
        img = np.zeros((5,5)).view(ma.MaskedArray)
        img[:] = 3*(np.arange(5.)-2)**2+2*(np.arange(5.)-2)
        img[2,2] = 1000
        bkgnd_params, img_mask = PSF.background_gradient_fit(img)
        self.assertFalse(np.sum((np.array(bkgnd_params)-
                                 np.array([0,2,0,3,0,0]))**2) < 1e-10)
        self.assertEqual(np.sum(img_mask), 0)
        bkgnd_params, img_mask = PSF.background_gradient_fit(img, ignore_center=1)
        self.assertTrue(np.sum((np.array(bkgnd_params)-
                                np.array([0,2,0,3,0,0]))**2) < 1e-10)
        self.assertEqual(np.sum(img_mask), 9)
        bkgnd_params, img_mask = PSF.background_gradient_fit(img, ignore_center=2)
        self.assertIsNone(bkgnd_params)
        self.assertIsNone(img_mask)
        bkgnd_params, img_mask = PSF.background_gradient_fit(img, ignore_center=3)
        self.assertIsNone(bkgnd_params)
        self.assertIsNone(img_mask)

        # Removal of bad pixels
        img = np.zeros((5,5)).view(ma.MaskedArray)
        img[:] = 3*(np.arange(5.)-2)**2+2*(np.arange(5.)-2)
        bkgnd_params, img_mask = PSF.background_gradient_fit(img, num_sigma=5)
        self.assertTrue(np.sum((np.array(bkgnd_params)-
                                np.array([0,2,0,3,0,0]))**2) < 1e-10)
        self.assertEqual(np.sum(img_mask), 0)
        img[2,2] = 10000
        bkgnd_params, img_mask = PSF.background_gradient_fit(img, num_sigma=3)
        self.assertTrue(np.sum((np.array(bkgnd_params)-
                                np.array([0,2,0,3,0,0]))**2) < 1e-10)
        self.assertEqual(np.sum(img_mask), 2)
        img[0,0] = 100000
        bkgnd_params, img_mask = PSF.background_gradient_fit(img, num_sigma=3)
        self.assertTrue(np.sum((np.array(bkgnd_params)-
                                np.array([0,2,0,3,0,0]))**2) < 1e-10)
        self.assertEqual(np.sum(img_mask), 2)
        img[0,4] = 10000
        bkgnd_params, img_mask = PSF.background_gradient_fit(img, num_sigma=3)
        self.assertTrue(np.sum((np.array(bkgnd_params)-
                                np.array([0,2,0,3,0,0]))**2) < 1e-10)
        self.assertEqual(np.sum(img_mask), 4)
